# Copyright 2016 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from charmhelpers.fetch import (
    apt_cache,
    apt_purge
)
from charmhelpers.core.hookenv import (
    log,
    DEBUG,
    WARNING,
)
from charmhelpers.contrib.hardening.audits import BaseAudit
from charmhelpers.fetch import ubuntu_apt_pkg as apt_pkg


class AptConfig(BaseAudit):

    def __init__(self, config, **kwargs):
        self.config = config

    def verify_config(self):
        apt_pkg.init()
        for cfg in self.config:
            value = apt_pkg.config.get(cfg['key'], cfg.get('default', ''))
            if value and value != cfg['expected']:
                log("APT config '%s' has unexpected value '%s' "
                    "(expected='%s')" %
                    (cfg['key'], value, cfg['expected']), level=WARNING)

    def ensure_compliance(self):
        self.verify_config()


class RestrictedPackages(BaseAudit):
    """Class used to audit restricted packages on the system."""

    def __init__(self, pkgs, **kwargs):
        super(RestrictedPackages, self).__init__(**kwargs)
        if isinstance(pkgs, str) or not hasattr(pkgs, '__iter__'):
            self.pkgs = pkgs.split()
        else:
            self.pkgs = pkgs

    def ensure_compliance(self):
        cache = apt_cache()

        for p in self.pkgs:
            if p not in cache:
                continue

            pkg = cache[p]
            if not self.is_virtual_package(pkg):
                if not pkg.current_ver:
                    log("Package '%s' is not installed." % pkg.name,
                        level=DEBUG)
                    continue
                else:
                    log("Restricted package '%s' is installed" % pkg.name,
                        level=WARNING)
                    self.delete_package(cache, pkg)
            else:
                log("Checking restricted virtual package '%s' provides" %
                    pkg.name, level=DEBUG)
                self.delete_package(cache, pkg)

    def delete_package(self, cache, pkg):
        """Deletes the package from the system.

        Deletes the package form the system, properly handling virtual
        packages.

        :param cache: the apt cache
        :param pkg: the package to remove
        """
        if self.is_virtual_package(pkg):
            log("Package '%s' appears to be virtual - purging provides" %
                pkg.name, level=DEBUG)
            for _p in pkg.provides_list:
                self.delete_package(cache, _p[2].parent_pkg)
        elif not pkg.current_ver:
            log("Package '%s' not installed" % pkg.name, level=DEBUG)
            return
        else:
            log("Purging package '%s'" % pkg.name, level=DEBUG)
            apt_purge(pkg.name)

    def is_virtual_package(self, pkg):
        return (pkg.get('has_provides', False) and
                not pkg.get('has_versions', False))



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

from charmhelpers.fetch import apt_install, apt_update
from charmhelpers.core.hookenv import (
    log,
    ERROR,
    INFO,
    TRACE
)
from charmhelpers.contrib.openstack.utils import OPENSTACK_CODENAMES

try:
    from jinja2 import FileSystemLoader, ChoiceLoader, Environment, exceptions
except ImportError:
    apt_update(fatal=True)
    apt_install('python3-jinja2', fatal=True)
    from jinja2 import FileSystemLoader, ChoiceLoader, Environment, exceptions


class OSConfigException(Exception):
    pass


def get_loader(templates_dir, os_release):
    """
    Create a jinja2.ChoiceLoader containing template dirs up to
    and including os_release.  If directory template directory
    is missing at templates_dir, it will be omitted from the loader.
    templates_dir is added to the bottom of the search list as a base
    loading dir.

    A charm may also ship a templates dir with this module
    and it will be appended to the bottom of the search list, eg::

        hooks/charmhelpers/contrib/openstack/templates

    :param templates_dir (str): Base template directory containing release
        sub-directories.
    :param os_release (str): OpenStack release codename to construct template
        loader.
    :returns: jinja2.ChoiceLoader constructed with a list of
        jinja2.FilesystemLoaders, ordered in descending
        order by OpenStack release.
    """
    tmpl_dirs = [(rel, os.path.join(templates_dir, rel))
                 for rel in OPENSTACK_CODENAMES.values()]

    if not os.path.isdir(templates_dir):
        log('Templates directory not found @ %s.' % templates_dir,
            level=ERROR)
        raise OSConfigException

    # the bottom contains tempaltes_dir and possibly a common templates dir
    # shipped with the helper.
    loaders = [FileSystemLoader(templates_dir)]
    helper_templates = os.path.join(os.path.dirname(__file__), 'templates')
    if os.path.isdir(helper_templates):
        loaders.append(FileSystemLoader(helper_templates))

    for rel, tmpl_dir in tmpl_dirs:
        if os.path.isdir(tmpl_dir):
            loaders.insert(0, FileSystemLoader(tmpl_dir))
        if rel == os_release:
            break
    # demote this log to the lowest level; we don't really need to see these
    # lots in production even when debugging.
    log('Creating choice loader with dirs: %s' %
        [l.searchpath for l in loaders], level=TRACE)
    return ChoiceLoader(loaders)


class OSConfigTemplate(object):
    """
    Associates a config file template with a list of context generators.
    Responsible for constructing a template context based on those generators.
    """

    def __init__(self, config_file, contexts, config_template=None):
        self.config_file = config_file

        if hasattr(contexts, '__call__'):
            self.contexts = [contexts]
        else:
            self.contexts = contexts

        self._complete_contexts = []

        self.config_template = config_template

    def context(self):
        ctxt = {}
        for context in self.contexts:
            _ctxt = context()
            if _ctxt:
                ctxt.update(_ctxt)
                # track interfaces for every complete context.
                [self._complete_contexts.append(interface)
                 for interface in context.interfaces
                 if interface not in self._complete_contexts]
        return ctxt

    def complete_contexts(self):
        '''
        Return a list of interfaces that have satisfied contexts.
        '''
        if self._complete_contexts:
            return self._complete_contexts
        self.context()
        return self._complete_contexts

    @property
    def is_string_template(self):
        """:returns: Boolean if this instance is a template initialised with a string"""
        return self.config_template is not None


class OSConfigRenderer(object):
    """
    This class provides a common templating system to be used by OpenStack
    charms.  It is intended to help charms share common code and templates,
    and ease the burden of managing config templates across multiple OpenStack
    releases.

    Basic usage::

        # import some common context generates from charmhelpers
        from charmhelpers.contrib.openstack import context

        # Create a renderer object for a specific OS release.
        configs = OSConfigRenderer(templates_dir='/tmp/templates',
                                   openstack_release='folsom')
        # register some config files with context generators.
        configs.register(config_file='/etc/nova/nova.conf',
                         contexts=[context.SharedDBContext(),
                                   context.AMQPContext()])
        configs.register(config_file='/etc/nova/api-paste.ini',
                         contexts=[context.IdentityServiceContext()])
        configs.register(config_file='/etc/haproxy/haproxy.conf',
                         contexts=[context.HAProxyContext()])
        configs.register(config_file='/etc/keystone/policy.d/extra.cfg',
                         contexts=[context.ExtraPolicyContext()
                                   context.KeystoneContext()],
                         config_template=hookenv.config('extra-policy'))
        # write out a single config
        configs.write('/etc/nova/nova.conf')
        # write out all registered configs
        configs.write_all()

    **OpenStack Releases and template loading**

    When the object is instantiated, it is associated with a specific OS
    release.  This dictates how the template loader will be constructed.

    The constructed loader attempts to load the template from several places
    in the following order:
    - from the most recent OS release-specific template dir (if one exists)
    - the base templates_dir
    - a template directory shipped in the charm with this helper file.

    For the example above, '/tmp/templates' contains the following structure::

        /tmp/templates/nova.conf
        /tmp/templates/api-paste.ini
        /tmp/templates/grizzly/api-paste.ini
        /tmp/templates/havana/api-paste.ini

    Since it was registered with the grizzly release, it first searches
    the grizzly directory for nova.conf, then the templates dir.

    When writing api-paste.ini, it will find the template in the grizzly
    directory.

    If the object were created with folsom, it would fall back to the
    base templates dir for its api-paste.ini template.

    This system should help manage changes in config files through
    openstack releases, allowing charms to fall back to the most recently
    updated config template for a given release

    The haproxy.conf, since it is not shipped in the templates dir, will
    be loaded from the module directory's template directory, eg
    $CHARM/hooks/charmhelpers/contrib/openstack/templates.  This allows
    us to ship common templates (haproxy, apache) with the helpers.

    **Context generators**

    Context generators are used to generate template contexts during hook
    execution.  Doing so may require inspecting service relations, charm
    config, etc.  When registered, a config file is associated with a list
    of generators.  When a template is rendered and written, all context
    generates are called in a chain to generate the context dictionary
    passed to the jinja2 template. See context.py for more info.
    """
    def __init__(self, templates_dir, openstack_release):
        if not os.path.isdir(templates_dir):
            log('Could not locate templates dir %s' % templates_dir,
                level=ERROR)
            raise OSConfigException

        self.templates_dir = templates_dir
        self.openstack_release = openstack_release
        self.templates = {}
        self._tmpl_env = None

        if None in [Environment, ChoiceLoader, FileSystemLoader]:
            # if this code is running, the object is created pre-install hook.
            # jinja2 shouldn't get touched until the module is reloaded on next
            # hook execution, with proper jinja2 bits successfully imported.
            apt_install('python3-jinja2')

    def register(self, config_file, contexts, config_template=None):
        """
        Register a config file with a list of context generators to be called
        during rendering.
        config_template can be used to load a template from a string instead of
        using template loaders and template files.
        :param config_file (str): a path where a config file will be rendered
        :param contexts (list): a list of context dictionaries with kv pairs
        :param config_template (str): an optional template string to use
        """
        self.templates[config_file] = OSConfigTemplate(
            config_file=config_file,
            contexts=contexts,
            config_template=config_template
        )
        log('Registered config file: {}'.format(config_file),
            level=INFO)

    def _get_tmpl_env(self):
        if not self._tmpl_env:
            loader = get_loader(self.templates_dir, self.openstack_release)
            self._tmpl_env = Environment(loader=loader)

    def _get_template(self, template):
        self._get_tmpl_env()
        template = self._tmpl_env.get_template(template)
        log('Loaded template from {}'.format(template.filename),
            level=INFO)
        return template

    def _get_template_from_string(self, ostmpl):
        '''
        Get a jinja2 template object from a string.
        :param ostmpl: OSConfigTemplate to use as a data source.
        '''
        self._get_tmpl_env()
        template = self._tmpl_env.from_string(ostmpl.config_template)
        log('Loaded a template from a string for {}'.format(
            ostmpl.config_file),
            level=INFO)
        return template

    def render(self, config_file):
        if config_file not in self.templates:
            log('Config not registered: {}'.format(config_file), level=ERROR)
            raise OSConfigException

        ostmpl = self.templates[config_file]
        ctxt = ostmpl.context()

        if ostmpl.is_string_template:
            template = self._get_template_from_string(ostmpl)
            log('Rendering from a string template: '
                '{}'.format(config_file),
                level=INFO)
        else:
            _tmpl = os.path.basename(config_file)
            try:
                template = self._get_template(_tmpl)
            except exceptions.TemplateNotFound:
                # if no template is found with basename, try looking
                # for it using a munged full path, eg:
                # /etc/apache2/apache2.conf -> etc_apache2_apache2.conf
                _tmpl = '_'.join(config_file.split('/')[1:])
                try:
                    template = self._get_template(_tmpl)
                except exceptions.TemplateNotFound as e:
                    log('Could not load template from {} by {} or {}.'
                        ''.format(
                            self.templates_dir,
                            os.path.basename(config_file),
                            _tmpl
                        ),
                        level=ERROR)
                    raise e

            log('Rendering from template: {}'.format(config_file),
                level=INFO)
        return template.render(ctxt)

    def write(self, config_file):
        """
        Write a single config file, raises if config file is not registered.
        """
        if config_file not in self.templates:
            log('Config not registered: %s' % config_file, level=ERROR)
            raise OSConfigException

        _out = self.render(config_file).encode('UTF-8')

        with open(config_file, 'wb') as out:
            out.write(_out)

        log('Wrote template %s.' % config_file, level=INFO)

    def write_all(self):
        """
        Write out all registered config files.
        """
        for k in self.templates.keys():
            self.write(k)

    def set_release(self, openstack_release):
        """
        Resets the template environment and generates a new template loader
        based on a the new openstack release.
        """
        self._tmpl_env = None
        self.openstack_release = openstack_release
        self._get_tmpl_env()

    def complete_contexts(self):
        '''
        Returns a list of context interfaces that yield a complete context.
        '''
        interfaces = []
        for i in self.templates.values():
            interfaces.extend(i.complete_contexts())
        return interfaces

    def get_incomplete_context_data(self, interfaces):
        '''
        Return dictionary of relation status of interfaces and any missing
        required context data. Example:
            {'amqp': {'missing_data': ['rabbitmq_password'], 'related': True},
             'zeromq-configuration': {'related': False}}
        '''
        incomplete_context_data = {}

        for i in self.templates.values():
            for context in i.contexts:
                for interface in interfaces:
                    related = False
                    if interface in context.interfaces:
                        related = context.get_related()
                        missing_data = context.missing_data
                        if missing_data:
                            incomplete_context_data[interface] = {'missing_data': missing_data}
                        if related:
                            if incomplete_context_data.get(interface):
                                incomplete_context_data[interface].update({'related': True})
                            else:
                                incomplete_context_data[interface] = {'related': True}
                        else:
                            incomplete_context_data[interface] = {'related': False}
        return incomplete_context_data



''' General helper functions for tests '''
from contextlib import contextmanager
from mock import patch, MagicMock
import io


@contextmanager
def patch_open():
    '''Patch open() to allow mocking both open() itself and the file that is
    yielded.

    Yields the mock for "open" and "file", respectively.'''
    mock_open = MagicMock(spec=open)
    mock_file = MagicMock(spec=io.FileIO)

    @contextmanager
    def stub_open(*args, **kwargs):
        mock_open(*args, **kwargs)
        yield mock_file

    with patch('builtins.open', stub_open):
        yield mock_open, mock_file


@contextmanager
def mock_open(filename, contents=None):
    ''' Slightly simpler mock of open to return contents for filename '''
    def mock_file(name, mode='r', buffering=-1):  # Python 2 signature.
        if name == filename:
            if 'b' in mode:
                return io.BytesIO(contents)
            return io.StringIO(contents)
        else:
            return open(name, mode, buffering)

    with patch('builtins.open', mock_file):
        yield


class FakeRelation(object):
    '''
    A fake relation class. Lets tests specify simple relation data
    for a default relation + unit (foo:0, foo/0, set in setUp()), eg:

        rel = {
            'private-address': 'foo',
            'password': 'passwd',
        }
        relation = FakeRelation(rel)
        self.relation_get.side_effect = relation.get
        passwd = self.relation_get('password')

    or more complex relations meant to be addressed by explicit relation id
    + unit id combos:

        rel = {
            'mysql:0': {
                'mysql/0': {
                    'private-address': 'foo',
                    'password': 'passwd',
                }
            }
        }
        relation = FakeRelation(rel)
        self.relation_get.side_affect = relation.get
        passwd = self.relation_get('password', rid='mysql:0', unit='mysql/0')

    set_relation_context can be used to simulate being in a relation hook
    context. This allows omitting a relation id or unit when calling relation
    helpers as the related unit is present.

    To set the context:

        relation = FakeRelation(rel)
        relation.set_relation_context('mysql-svc2/0', 'shared-db:12')

    To clear it:

        relation.clear_relation_context()
    '''
    def __init__(self, relation_data):
        self.relation_data = relation_data
        self.remote_unit = None
        self.current_relation_id = None

    def set_relation_context(self, remote_unit, relation_id):
        self.remote_unit = remote_unit
        self.current_relation_id = relation_id

    def clear_relation_context(self):
        self.remote_unit = None
        self.current_relation_id = None

    def get(self, attribute=None, unit=None, rid=None):
        if not rid or rid == 'foo:0':
            if self.current_relation_id:
                if not unit:
                    unit = self.remote_unit
                udata = self.relation_data[self.current_relation_id][unit]
                if attribute:
                    return udata[attribute]
                else:
                    return udata[unit]
            if attribute is None:
                return self.relation_data
            elif attribute in self.relation_data:
                return self.relation_data[attribute]
            return None
        else:
            if rid not in self.relation_data:
                return None
            try:
                relation = self.relation_data[rid][unit]
            except KeyError:
                return None
            if attribute and attribute in relation:
                return relation[attribute]
            return relation

    def relation_id(self):
        return self.current_relation_id

    def relation_ids(self, reltype=None):
        rids = self.relation_data.keys()
        if reltype:
            return [r for r in rids if r.split(':')[0] == reltype]
        return rids

    def related_units(self, relid=None):
        try:
            return self.relation_data[relid].keys()
        except KeyError:
            return []

    def relation_units(self, relation_id):
        if relation_id not in self.relation_data:
            return None
        return self.relation_data[relation_id].keys()



#!/usr/bin/python3

# Copyright 2014-2022 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Checks for services with deferred restarts.

This Nagios check will parse /var/lib/policy-rd.d/
to find any restarts that are currently deferred.
"""

import argparse
import glob
import sys
import yaml


DEFERRED_EVENTS_DIR = '/var/lib/policy-rc.d'


def get_deferred_events():
    """Return a list of deferred events dicts from policy-rc.d files.

    Events are read from DEFERRED_EVENTS_DIR and are of the form:
    {
        action: restart,
        policy_requestor_name: rabbitmq-server,
        policy_requestor_type: charm,
        reason: 'Pkg update',
        service: rabbitmq-server,
        time: 1614328743
    }

    :raises OSError: Raised in case of a system error while reading a policy file
    :raises yaml.YAMLError: Raised if parsing a policy file fails

    :returns: List of deferred event dictionaries
    :rtype: list
    """
    deferred_events_files = glob.glob(
        '{}/*.deferred'.format(DEFERRED_EVENTS_DIR))

    deferred_events = []
    for event_file in deferred_events_files:
        with open(event_file, 'r') as f:
            event = yaml.safe_load(f)
            deferred_events.append(event)

    return deferred_events


def get_deferred_restart_services(application=None):
    """Returns a list of services with deferred restarts.

    :param str application: Name of the application that blocked the service restart.
                            If application is None, all services with deferred restarts
                            are returned. Services which are blocked by a non-charm
                            requestor are always returned.

    :raises OSError: Raised in case of a system error while reading a policy file
    :raises yaml.YAMLError: Raised if parsing a policy file fails

    :returns: List of services with deferred restarts belonging to application.
    :rtype: list
    """

    deferred_restart_events = filter(
        lambda e: e['action'] == 'restart', get_deferred_events())

    deferred_restart_services = set()
    for restart_event in deferred_restart_events:
        if application:
            if (
                restart_event['policy_requestor_type'] != 'charm' or
                restart_event['policy_requestor_type'] == 'charm' and
                restart_event['policy_requestor_name'] == application
            ):
                deferred_restart_services.add(restart_event['service'])
        else:
            deferred_restart_services.add(restart_event['service'])

    return list(deferred_restart_services)


def main():
    """Check for services with deferred restarts."""
    parser = argparse.ArgumentParser(
        description='Check for services with deferred restarts')
    parser.add_argument(
        '--application', help='Check services belonging to this application only')

    args = parser.parse_args()

    services = set(get_deferred_restart_services(args.application))

    if len(services) == 0:
        print('OK: No deferred service restarts.')
        sys.exit(0)
    else:
        print(
            'CRITICAL: Restarts are deferred for services: {}.'.format(', '.join(services)))
        sys.exit(1)


if __name__ == '__main__':
    try:
        main()
    except OSError as e:
        print('CRITICAL: A system error occurred: {} ({})'.format(e.errno, e.strerror))
        sys.exit(1)
    except yaml.YAMLError as e:
        print('CRITICAL: Failed to parse a policy file: {}'.format(str(e)))
        sys.exit(1)
    except Exception as e:
        print('CRITICAL: An unknown error occurred: {}'.format(str(e)))
        sys.exit(1)



# Copyright 2019 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenStack Security Audit code"""

import collections
from enum import Enum
import traceback

from charmhelpers.core.host import cmp_pkgrevno
import charmhelpers.contrib.openstack.utils as openstack_utils
import charmhelpers.core.hookenv as hookenv


class AuditType(Enum):
    OpenStackSecurityGuide = 1


_audits = {}

Audit = collections.namedtuple('Audit', 'func filters')


def audit(*args):
    """Decorator to register an audit.

    These are used to generate audits that can be run on a
    deployed system that matches the given configuration

    :param args: List of functions to filter tests against
    :type args: List[Callable[Dict]]
    """
    def wrapper(f):
        test_name = f.__name__
        if _audits.get(test_name):
            raise RuntimeError(
                "Test name '{}' used more than once"
                .format(test_name))
        non_callables = [fn for fn in args if not callable(fn)]
        if non_callables:
            raise RuntimeError(
                "Configuration includes non-callable filters: {}"
                .format(non_callables))
        _audits[test_name] = Audit(func=f, filters=args)
        return f
    return wrapper


def is_audit_type(*args):
    """This audit is included in the specified kinds of audits.

    :param *args: List of AuditTypes to include this audit in
    :type args: List[AuditType]
    :rtype: Callable[Dict]
    """
    def _is_audit_type(audit_options):
        if audit_options.get('audit_type') in args:
            return True
        else:
            return False
    return _is_audit_type


def since_package(pkg, pkg_version):
    """This audit should be run after the specified package version (incl).

    :param pkg: Package name to compare
    :type pkg: str
    :param release: The package version
    :type release: str
    :rtype: Callable[Dict]
    """
    def _since_package(audit_options=None):
        return cmp_pkgrevno(pkg, pkg_version) >= 0

    return _since_package


def before_package(pkg, pkg_version):
    """This audit should be run before the specified package version (excl).

    :param pkg: Package name to compare
    :type pkg: str
    :param release: The package version
    :type release: str
    :rtype: Callable[Dict]
    """
    def _before_package(audit_options=None):
        return not since_package(pkg, pkg_version)()

    return _before_package


def since_openstack_release(pkg, release):
    """This audit should run after the specified OpenStack version (incl).

    :param pkg: Package name to compare
    :type pkg: str
    :param release: The OpenStack release codename
    :type release: str
    :rtype: Callable[Dict]
    """
    def _since_openstack_release(audit_options=None):
        _release = openstack_utils.get_os_codename_package(pkg)
        return openstack_utils.CompareOpenStackReleases(_release) >= release

    return _since_openstack_release


def before_openstack_release(pkg, release):
    """This audit should run before the specified OpenStack version (excl).

    :param pkg: Package name to compare
    :type pkg: str
    :param release: The OpenStack release codename
    :type release: str
    :rtype: Callable[Dict]
    """
    def _before_openstack_release(audit_options=None):
        return not since_openstack_release(pkg, release)()

    return _before_openstack_release


def it_has_config(config_key):
    """This audit should be run based on specified config keys.

    :param config_key: Config key to look for
    :type config_key: str
    :rtype: Callable[Dict]
    """
    def _it_has_config(audit_options):
        return audit_options.get(config_key) is not None

    return _it_has_config


def run(audit_options):
    """Run the configured audits with the specified audit_options.

    :param audit_options: Configuration for the audit
    :type audit_options: Config

    :rtype: Dict[str, str]
    """
    errors = {}
    results = {}
    for name, audit in sorted(_audits.items()):
        result_name = name.replace('_', '-')
        if result_name in audit_options.get('excludes', []):
            print(
                "Skipping {} because it is"
                "excluded in audit config"
                .format(result_name))
            continue
        if all(p(audit_options) for p in audit.filters):
            try:
                audit.func(audit_options)
                print("{}: PASS".format(name))
                results[result_name] = {
                    'success': True,
                }
            except AssertionError as e:
                print("{}: FAIL ({})".format(name, e))
                results[result_name] = {
                    'success': False,
                    'message': e,
                }
            except Exception as e:
                print("{}: ERROR ({})".format(name, e))
                errors[name] = e
                results[result_name] = {
                    'success': False,
                    'message': e,
                }
    for name, error in errors.items():
        print("=" * 20)
        print("Error in {}: ".format(name))
        traceback.print_tb(error.__traceback__)
        print()
    return results


def action_parse_results(result):
    """Parse the result of `run` in the context of an action.

    :param result: The result of running the security-checklist
        action on a unit
    :type result: Dict[str, Dict[str, str]]
    :rtype: int
    """
    passed = True
    for test, result in result.items():
        if result['success']:
            hookenv.action_set({test: 'PASS'})
        else:
            hookenv.action_set({test: 'FAIL - {}'.format(result['message'])})
            passed = False
    if not passed:
        hookenv.action_fail("One or more tests failed")
    return 0 if passed else 1



# Copyright 2016 Canonical Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
LSPCI = b"""
0000:00:00.0 "Host bridge" "Intel Corporation" "Haswell-E DMI2" -r02 "Intel Corporation" "Device 0000"
0000:00:03.0 "PCI bridge" "Intel Corporation" "Haswell-E PCI Express Root Port 3" -r02 "" ""
0000:00:03.2 "PCI bridge" "Intel Corporation" "Haswell-E PCI Express Root Port 3" -r02 "" ""
0000:00:05.0 "System peripheral" "Intel Corporation" "Haswell-E Address Map, VTd_Misc, System Management" -r02 "" ""
0000:00:05.1 "System peripheral" "Intel Corporation" "Haswell-E Hot Plug" -r02 "" ""
0000:00:05.2 "System peripheral" "Intel Corporation" "Haswell-E RAS, Control Status and Global Errors" -r02 "" ""
0000:00:05.4 "PIC" "Intel Corporation" "Haswell-E I/O Apic" -r02 -p20 "Intel Corporation" "Device 0000"
0000:00:11.0 "Unassigned class [ff00]" "Intel Corporation" "Wellsburg SPSR" -r05 "Intel Corporation" "Device 7270"
0000:00:11.4 "SATA controller" "Intel Corporation" "Wellsburg sSATA Controller [AHCI mode]" -r05 -p01 "Cisco Systems Inc" "Device 0067"
0000:00:16.0 "Communication controller" "Intel Corporation" "Wellsburg MEI Controller #1" -r05 "Intel Corporation" "Device 7270"
0000:00:16.1 "Communication controller" "Intel Corporation" "Wellsburg MEI Controller #2" -r05 "Intel Corporation" "Device 7270"
0000:00:1a.0 "USB controller" "Intel Corporation" "Wellsburg USB Enhanced Host Controller #2" -r05 -p20 "Intel Corporation" "Device 7270"
0000:00:1c.0 "PCI bridge" "Intel Corporation" "Wellsburg PCI Express Root Port #1" -rd5 "" ""
0000:00:1c.3 "PCI bridge" "Intel Corporation" "Wellsburg PCI Express Root Port #4" -rd5 "" ""
0000:00:1c.4 "PCI bridge" "Intel Corporation" "Wellsburg PCI Express Root Port #5" -rd5 "" ""
0000:00:1d.0 "USB controller" "Intel Corporation" "Wellsburg USB Enhanced Host Controller #1" -r05 -p20 "Intel Corporation" "Device 7270"
0000:00:1f.0 "ISA bridge" "Intel Corporation" "Wellsburg LPC Controller" -r05 "Intel Corporation" "Device 7270"
0000:00:1f.2 "SATA controller" "Intel Corporation" "Wellsburg 6-Port SATA Controller [AHCI mode]" -r05 -p01 "Cisco Systems Inc" "Device 0067"
0000:01:00.0 "PCI bridge" "Cisco Systems Inc" "VIC 82 PCIe Upstream Port" -r01 "" ""
0000:02:00.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:02:01.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:03:00.0 "Unclassified device [00ff]" "Cisco Systems Inc" "VIC Management Controller" -ra2 "Cisco Systems Inc" "Device 012e"
0000:04:00.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Upstream Port" -ra2 "" ""
0000:05:00.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:05:01.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:05:02.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:05:03.0 "PCI bridge" "Cisco Systems Inc" "VIC PCIe Downstream Port" -ra2 "" ""
0000:08:00.0 "Fibre Channel" "Cisco Systems Inc" "VIC FCoE HBA" -ra2 "Cisco Systems Inc" "Device 012e"
0000:09:00.0 "Fibre Channel" "Cisco Systems Inc" "VIC FCoE HBA" -ra2 "Cisco Systems Inc" "Device 012e"
0000:0b:00.0 "RAID bus controller" "LSI Logic / Symbios Logic" "MegaRAID SAS-3 3108 [Invader]" -r02 "Cisco Systems Inc" "Device 00db"
0000:0f:00.0 "VGA compatible controller" "Matrox Electronics Systems Ltd." "MGA G200e [Pilot] ServerEngines (SEP1)" -r02 "Cisco Systems Inc" "Device 0101"
0000:10:00.0 "Ethernet controller" "Intel Corporation" "I350 Gigabit Network Connection" -r01 "Cisco Systems Inc" "Device 00d6"
0000:10:00.1 "Ethernet controller" "Intel Corporation" "I350 Gigabit Network Connection" -r01 "Cisco Systems Inc" "Device 00d6"
0000:7f:08.0 "System peripheral" "Intel Corporation" "Haswell-E QPI Link 0" -r02 "Intel Corporation" "Haswell-E QPI Link 0"
"""

SYS_TREE = {
    "/sys/class/net/eth2": "../../devices/pci0000:00/0000:00:1c.4/0000:10:00.0/net/eth2",
    "/sys/class/net/eth3": "../../devices/pci0000:00/0000:00:1c.4/0000:10:00.1/net/eth3",
    "/sys/class/net/juju-br0": "../../devices/virtual/net/juju-br0",
    "/sys/class/net/lo": "../../devices/virtual/net/lo",
    "/sys/class/net/lxcbr0": "../../devices/virtual/net/lxcbr0",
    "/sys/class/net/veth1GVRCF": "../../devices/virtual/net/veth1GVRCF",
    "/sys/class/net/veth7AXEUK": "../../devices/virtual/net/veth7AXEUK",
    "/sys/class/net/vethACOIJJ": "../../devices/virtual/net/vethACOIJJ",
    "/sys/class/net/vethMQ819H": "../../devices/virtual/net/vethMQ819H",
    "/sys/class/net/virbr0": "../../devices/virtual/net/virbr0",
    "/sys/class/net/virbr0-nic": "../../devices/virtual/net/virbr0-nic",
    "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.0/net/eth2/device": "../../../0000:10:00.0",
    "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.1/net/eth3/device": "../../../0000:10:00.1",
    "/sys/bus/pci/devices/0000:10:00.0": "../../../devices/pci0000:00/0000:00:1c.4/0000:10:00.0",
    "/sys/bus/pci/devices/0000:10:00.1": "../../../devices/pci0000:00/0000:00:1c.4/0000:10:00.1",
    "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.0": "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.0",
    "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.1": "/sys/devices/pci0000:00/0000:00:1c.4/0000:10:00.1",
}

NETDEVS = {
    "/sys/bus/pci/devices/0000:10:00.0/net": ["eth2"],
    # list the pf last so we validate the code successfully traversing the
    # non-pf netdevs
    "/sys/bus/pci/devices/0000:10:00.1/net": ["eth33", "eth51", "eth3"],
}

FILE_CONTENTS = {
    "/sys/bus/pci/devices/0000:10:00.0/net/eth2/address": "a8:9d:21:cf:93:fc",
    "/sys/bus/pci/devices/0000:10:00.1/net/eth3/address": "a8:9d:21:cf:93:fd",
    "/sys/bus/pci/devices/0000:10:00.0/net/eth2/operstate": "up",
    "/sys/bus/pci/devices/0000:10:00.1/net/eth3/operstate": "down",
    "/sys/bus/pci/devices/0000:10:00.1/net/eth3/phys_port_name": "p0",
    "/sys/bus/pci/devices/0000:10:00.1/net/eth33/phys_port_name": "notapf",
    # eth51 deliberately not listed to validate non-presence of or failure to
    # open phys_port_name
}

COMMANDS = {"LSPCI_MD": ["lspci", "-m", "-D"]}

NET_SETUP = {"LSPCI_MD": LSPCI}



from testtools import TestCase
from mock import patch

import charmhelpers.contrib.openstack.alternatives as alternatives


NAME = 'test'
SOURCE = '/var/lib/charm/test/test.conf'
TARGET = '/etc/test/test,conf'


class AlternativesTestCase(TestCase):

    @patch('subprocess.os.path')
    @patch('subprocess.check_call')
    def test_new_alternative(self, _check, _path):
        _path.exists.return_value = False
        alternatives.install_alternative(NAME,
                                         TARGET,
                                         SOURCE)
        _check.assert_called_with(
            ['update-alternatives', '--force', '--install',
             TARGET, NAME, SOURCE, '50']
        )

    @patch('subprocess.os.path')
    @patch('subprocess.check_call')
    def test_priority(self, _check, _path):
        _path.exists.return_value = False
        alternatives.install_alternative(NAME,
                                         TARGET,
                                         SOURCE, 100)
        _check.assert_called_with(
            ['update-alternatives', '--force', '--install',
             TARGET, NAME, SOURCE, '100']
        )

    @patch('shutil.move')
    @patch('subprocess.os.path')
    @patch('subprocess.check_call')
    def test_new_alternative_existing_file(self, _check,
                                           _path, _move):
        _path.exists.return_value = True
        _path.islink.return_value = False
        alternatives.install_alternative(NAME,
                                         TARGET,
                                         SOURCE)
        _check.assert_called_with(
            ['update-alternatives', '--force', '--install',
             TARGET, NAME, SOURCE, '50']
        )
        _move.assert_called_with(TARGET, '{}.bak'.format(TARGET))

    @patch('shutil.move')
    @patch('subprocess.os.path')
    @patch('subprocess.check_call')
    def test_new_alternative_existing_link(self, _check,
                                           _path, _move):
        _path.exists.return_value = True
        _path.islink.return_value = True
        alternatives.install_alternative(NAME,
                                         TARGET,
                                         SOURCE)
        _check.assert_called_with(
            ['update-alternatives', '--force', '--install',
             TARGET, NAME, SOURCE, '50']
        )
        _move.assert_not_called()

    @patch('subprocess.check_call')
    def test_remove_alternative(self, _check):
        alternatives.remove_alternative(NAME, SOURCE)
        _check.assert_called_with(
            ['update-alternatives', '--remove',
             NAME, SOURCE]
        )



# Copyright 2014-2015 Canonical Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import yaml

from charmhelpers.core import hookenv
from charmhelpers.core import host
from charmhelpers.core import templating

from charmhelpers.core.services.base import ManagerCallback


__all__ = ['RelationContext', 'TemplateCallback',
           'render_template', 'template']


class RelationContext(dict):
    """
    Base class for a context generator that gets relation data from juju.

    Subclasses must provide the attributes `name`, which is the name of the
    interface of interest, `interface`, which is the type of the interface of
    interest, and `required_keys`, which is the set of keys required for the
    relation to be considered complete.  The data for all interfaces matching
    the `name` attribute that are complete will used to populate the dictionary
    values (see `get_data`, below).

    The generated context will be namespaced under the relation :attr:`name`,
    to prevent potential naming conflicts.

    :param str name: Override the relation :attr:`name`, since it can vary from charm to charm
    :param list additional_required_keys: Extend the list of :attr:`required_keys`
    """
    name = None
    interface = None

    def __init__(self, name=None, additional_required_keys=None):
        if not hasattr(self, 'required_keys'):
            self.required_keys = []

        if name is not None:
            self.name = name
        if additional_required_keys:
            self.required_keys.extend(additional_required_keys)
        self.get_data()

    def __bool__(self):
        """
        Returns True if all of the required_keys are available.
        """
        return self.is_ready()

    __nonzero__ = __bool__

    def __repr__(self):
        return super(RelationContext, self).__repr__()

    def is_ready(self):
        """
        Returns True if all of the `required_keys` are available from any units.
        """
        ready = len(self.get(self.name, [])) > 0
        if not ready:
            hookenv.log('Incomplete relation: {}'.format(self.__class__.__name__), hookenv.DEBUG)
        return ready

    def _is_ready(self, unit_data):
        """
        Helper method that tests a set of relation data and returns True if
        all of the `required_keys` are present.
        """
        return set(unit_data.keys()).issuperset(set(self.required_keys))

    def get_data(self):
        """
        Retrieve the relation data for each unit involved in a relation and,
        if complete, store it in a list under `self[self.name]`.  This
        is automatically called when the RelationContext is instantiated.

        The units are sorted lexographically first by the service ID, then by
        the unit ID.  Thus, if an interface has two other services, 'db:1'
        and 'db:2', with 'db:1' having two units, 'wordpress/0' and 'wordpress/1',
        and 'db:2' having one unit, 'mediawiki/0', all of which have a complete
        set of data, the relation data for the units will be stored in the
        order: 'wordpress/0', 'wordpress/1', 'mediawiki/0'.

        If you only care about a single unit on the relation, you can just
        access it as `{{ interface[0]['key'] }}`.  However, if you can at all
        support multiple units on a relation, you should iterate over the list,
        like::

            {% for unit in interface -%}
                {{ unit['key'] }}{% if not loop.last %},{% endif %}
            {%- endfor %}

        Note that since all sets of relation data from all related services and
        units are in a single list, if you need to know which service or unit a
        set of data came from, you'll need to extend this class to preserve
        that information.
        """
        if not hookenv.relation_ids(self.name):
            return

        ns = self.setdefault(self.name, [])
        for rid in sorted(hookenv.relation_ids(self.name)):
            for unit in sorted(hookenv.related_units(rid)):
                reldata = hookenv.relation_get(rid=rid, unit=unit)
                if self._is_ready(reldata):
                    ns.append(reldata)

    def provide_data(self):
        """
        Return data to be relation_set for this interface.
        """
        return {}


class MysqlRelation(RelationContext):
    """
    Relation context for the `mysql` interface.

    :param str name: Override the relation :attr:`name`, since it can vary from charm to charm
    :param list additional_required_keys: Extend the list of :attr:`required_keys`
    """
    name = 'db'
    interface = 'mysql'

    def __init__(self, *args, **kwargs):
        self.required_keys = ['host', 'user', 'password', 'database']
        RelationContext.__init__(self, *args, **kwargs)


class HttpRelation(RelationContext):
    """
    Relation context for the `http` interface.

    :param str name: Override the relation :attr:`name`, since it can vary from charm to charm
    :param list additional_required_keys: Extend the list of :attr:`required_keys`
    """
    name = 'website'
    interface = 'http'

    def __init__(self, *args, **kwargs):
        self.required_keys = ['host', 'port']
        RelationContext.__init__(self, *args, **kwargs)

    def provide_data(self):
        return {
            'host': hookenv.unit_get('private-address'),
            'port': 80,
        }


class RequiredConfig(dict):
    """
    Data context that loads config options with one or more mandatory options.

    Once the required options have been changed from their default values, all
    config options will be available, namespaced under `config` to prevent
    potential naming conflicts (for example, between a config option and a
    relation property).

    :param list *args: List of options that must be changed from their default values.
    """

    def __init__(self, *args):
        self.required_options = args
        self['config'] = hookenv.config()
        with open(os.path.join(hookenv.charm_dir(), 'config.yaml')) as fp:
            self.config = yaml.safe_load(fp).get('options', {})

    def __bool__(self):
        for option in self.required_options:
            if option not in self['config']:
                return False
            current_value = self['config'][option]
            default_value = self.config[option].get('default')
            if current_value == default_value:
                return False
            if current_value in (None, '') and default_value in (None, ''):
                return False
        return True

    def __nonzero__(self):
        return self.__bool__()


class StoredContext(dict):
    """
    A data context that always returns the data that it was first created with.

    This is useful to do a one-time generation of things like passwords, that
    will thereafter use the same value that was originally generated, instead
    of generating a new value each time it is run.
    """
    def __init__(self, file_name, config_data):
        """
        If the file exists, populate `self` with the data from the file.
        Otherwise, populate with the given data and persist it to the file.
        """
        if os.path.exists(file_name):
            self.update(self.read_context(file_name))
        else:
            self.store_context(file_name, config_data)
            self.update(config_data)

    def store_context(self, file_name, config_data):
        if not os.path.isabs(file_name):
            file_name = os.path.join(hookenv.charm_dir(), file_name)
        with open(file_name, 'w') as file_stream:
            os.fchmod(file_stream.fileno(), 0o600)
            yaml.dump(config_data, file_stream)

    def read_context(self, file_name):
        if not os.path.isabs(file_name):
            file_name = os.path.join(hookenv.charm_dir(), file_name)
        with open(file_name, 'r') as file_stream:
            data = yaml.safe_load(file_stream)
            if not data:
                raise OSError("%s is empty" % file_name)
            return data


class TemplateCallback(ManagerCallback):
    """
    Callback class that will render a Jinja2 template, for use as a ready
    action.

    :param str source: The template source file, relative to
        `$CHARM_DIR/templates`

    :param str target: The target to write the rendered template to (or None)
    :param str owner: The owner of the rendered file
    :param str group: The group of the rendered file
    :param int perms: The permissions of the rendered file
    :param partial on_change_action: functools partial to be executed when
                                     rendered file changes
    :param jinja2 loader template_loader: A jinja2 template loader

    :return str: The rendered template
    """
    def __init__(self, source, target,
                 owner='root', group='root', perms=0o444,
                 on_change_action=None, template_loader=None):
        self.source = source
        self.target = target
        self.owner = owner
        self.group = group
        self.perms = perms
        self.on_change_action = on_change_action
        self.template_loader = template_loader

    def __call__(self, manager, service_name, event_name):
        pre_checksum = ''
        if self.on_change_action and os.path.isfile(self.target):
            pre_checksum = host.file_hash(self.target)
        service = manager.get_service(service_name)
        context = {'ctx': {}}
        for ctx in service.get('required_data', []):
            context.update(ctx)
            context['ctx'].update(ctx)

        result = templating.render(self.source, self.target, context,
                                   self.owner, self.group, self.perms,
                                   template_loader=self.template_loader)
        if self.on_change_action:
            if pre_checksum == host.file_hash(self.target):
                hookenv.log(
                    'No change detected: {}'.format(self.target),
                    hookenv.DEBUG)
            else:
                self.on_change_action()

        return result


# Convenience aliases for templates
render_template = template = TemplateCallback
